from nltk import pos_tag, word_tokenize
from nltk.chunk import ChunkParserI
from nltk.chunk import RegexpParser
from nltk.chunk import conlltags2tree
from nltk.corpus import gazetteers

from JarbasModelZoo import load_model
from JarbasModelZoo.features import extract_iob_features


class NamedEntityChunker(ChunkParserI):
    def __init__(self, model_id):
        super().__init__()
        self.feature_detector = extract_iob_features
        self.tagger = load_model(model_id)

    def parse(self, tagged_sent):
        chunks = self.tag(tagged_sent)

        # Transform the result from [((w1, t1), iob1), ...]
        # to the preferred list of triplets format [(w1, t1, iob1), ...]
        iob_triplets = [(w, t, c) for ((w, t), c) in chunks]

        # Transform the list of triplets to nltk.Tree format
        return conlltags2tree(iob_triplets)

    def tag(self, tagged_sent):
        if isinstance(tagged_sent, str):
            tagged_sent = pos_tag(word_tokenize(tagged_sent))

        return self.tagger.tag(tagged_sent)


class GazetteerLocationChunker(ChunkParserI):
    def __init__(self):
        self.locations = set(gazetteers.words())
        longest_word = max(self.locations, key=len)
        self.lookahead = len(longest_word.split(" "))

    def iob_locations(self, tagged_sent):
        if isinstance(tagged_sent, str):
            tagged_sent = pos_tag(word_tokenize(tagged_sent))

        found_locs = []
        for idx, (word, tag) in enumerate(tagged_sent):
            nexttags = tagged_sent[idx + 1:idx + self.lookahead]
            for i in range(len(nexttags)):
                loc_tags = nexttags[:i]
                loc_str = " ".join([word] + [n[0] for n in loc_tags])

                if loc_str in self.locations:
                    found_locs.append(idx)
                    yield word, tag, 'B-LOCATION'
                    for i, (w, t) in enumerate(loc_tags):
                        found_locs.append(i + idx + 1)
                        yield w, t, 'I-LOCATION'

                    break
            else:
                if word in self.locations:
                    found_locs.append(idx)
                    yield word, tag, 'B-LOCATION'

            if idx not in found_locs:
                yield word, tag, 'O'

    def tag(self, tagged_sent):
        return list(self.iob_locations(tagged_sent))

    def parse(self, tagged_sent):
        iobs = self.iob_locations(tagged_sent)
        return conlltags2tree(iobs)


class ProperNounChunker(ChunkParserI):
    def __init__(self):
        super().__init__()
        self.chunker = RegexpParser(r'''NAME:{<NNP>+}''')

    def parse(self, tagged_sent):
        if isinstance(tagged_sent, str):
            tagged_sent = pos_tag(word_tokenize(tagged_sent))
        return self.chunker.parse(tagged_sent)


class PartialRegexChunker(ChunkParserI):
    def __init__(self):
        super().__init__()
        self.chunker = RegexpParser(r'''
        NP:
            {<DT>?<NN.*>+}	# chunk optional determiner with nouns
            <JJ>{}<NN.*>	# merge adjective with noun chunk
        PP:
            {<IN>}			# chunk preposition
        VP: 
            {<MD>?<VB.*>}	# chunk optional modal with verb''')

    def parse(self, tagged_sent):
        if isinstance(tagged_sent, str):
            tagged_sent = pos_tag(word_tokenize(tagged_sent))
        return self.chunker.parse(tagged_sent)


class PostagChunkParser(ChunkParserI):
    def __init__(self, model_id, pos_tagger=None):
        self.chunker = load_model(model_id)
        if pos_tagger:
            self.tagger = load_model(pos_tagger)
        else:
            self.tagger = None

    def parse(self, sentence):
        if isinstance(sentence, str):
            if self.tagger:
                sentence = self.tagger.tag(word_tokenize(sentence))
            else:
                sentence = pos_tag(word_tokenize(sentence))
        pos_tags = [pos for word, pos in sentence]

        # Get the Chunk tags
        tagged_pos_tags = self.chunker.tag(pos_tags)

        # Assemble the (word, pos, chunk) triplets
        iob_triplets = [(word, pos_tag, chunk_tag)
                        for ((word, pos_tag), (pos_tag, chunk_tag)) in
                        zip(sentence, tagged_pos_tags)]

        # Transform the list of triplets to nltk.Tree format
        return conlltags2tree(iob_triplets)


class ClassifierChunkParser(ChunkParserI):
    def __init__(self, model_id, pos_tagger=None):
        self.chunker = load_model(model_id)
        if pos_tagger:
            self.tagger = load_model(pos_tagger)
        else:
            self.tagger = None

    def parse(self, tagged_sent):
        if isinstance(tagged_sent, str):
            if self.tagger:
                tagged_sent = self.tagger.tag(word_tokenize(tagged_sent))
            else:
                tagged_sent = pos_tag(word_tokenize(tagged_sent))
        chunks = self.chunker.tag(tagged_sent)

        # Transform the result from [((w1, t1), iob1), ...]
        # to the preferred list of triplets format [(w1, t1, iob1), ...]
        iob_triplets = [(w, t, c) for ((w, t), c) in chunks]

        # Transform the list of triplets to nltk.Tree format
        return conlltags2tree(iob_triplets)



# this file contains feature extractor functions used by some of the models
# this means that models using this require this package installed and can
# NOT be loaded directly with pickle module only
import re
import nltk
from nltk.stem.snowball import SnowballStemmer


def get_word_shape(word):
    word_shape = 'other'
    if re.match('[0-9]+(\.[0-9]*)?|[0-9]*\.[0-9]+$', word):
        word_shape = 'number'
    elif re.match('\W+$', word):
        word_shape = 'punct'
    elif re.match('[A-Z][a-z]+$', word):
        word_shape = 'capitalized'
    elif re.match('[A-Z]+$', word):
        word_shape = 'uppercase'
    elif re.match('[a-z]+$', word):
        word_shape = 'lowercase'
    elif re.match('[A-Z][a-z]+[A-Z][a-z]+[A-Za-z]*$', word):
        word_shape = 'camelcase'
    elif re.match('[A-Za-z]+$', word):
        word_shape = 'mixedcase'
    elif re.match('__.+__$', word):
        word_shape = 'wildcard'
    elif re.match('[A-Za-z0-9]+\.$', word):
        word_shape = 'ending-dot'
    elif re.match('[A-Za-z0-9]+\.[A-Za-z0-9\.]+\.$', word):
        word_shape = 'abbreviation'
    elif re.match('[A-Za-z0-9]+\-[A-Za-z0-9\-]+.*$', word):
        word_shape = 'contains-hyphen'

    return word_shape


def extract_iob_features(tokens, index, history, stemmer=None, memory=2):
    """
    `tokens`  = a POS-tagged sentence [(w1, t1), ...]
    `index`   = the index of the token we want to extract features for
    `history` = the previous predicted IOB tags
    """
    feat_dict = extract_postag_features(tokens, index, stemmer=stemmer,
                                        memory=memory)

    # Pad the sequence with placeholders
    tokens = ['O'] * memory + history

    index += memory

    # look back N predictions
    for i in range(1, memory + 1):
        k = "prev-" * i
        previob = tokens[index - i]
        # update with IOB features
        feat_dict[k + "iob"] = previob

    return feat_dict


def extract_postag_features(tokens, index, stemmer=None, memory=2):
    """
    `tokens`  = a POS-tagged sentence [(w1, t1), ...]
    `index`   = the index of the token we want to extract features for
    """
    original_toks = list(tokens)
    # word features
    feat_dict = extract_word_features([t[0] for t in tokens],
                                      index, stemmer, memory=memory)

    # Pad the sequence with placeholders
    tokens = []
    for i in range(1, memory + 1):
        tokens.append((f'__START{i}__', f'__START{i}__'))
    tokens = list(reversed(tokens)) + original_toks
    for i in range(1, memory + 1):
        tokens.append((f'__END{i}__', f'__END{i}__'))

    # shift the index to accommodate the padding
    index += memory

    word, pos = tokens[index]

    # update with postag features
    feat_dict["pos"] = pos

    # look ahead N words
    for i in range(1, memory + 1):
        k = "next-" * i
        nextword, nextpos = tokens[index + i]
        feat_dict[k + "pos"] = nextpos

    # look back N words
    for i in range(1, memory + 1):
        k = "prev-" * i
        prevword, prevpos = tokens[index - i]
        feat_dict[k + "pos"] = prevpos

    return feat_dict


def extract_word_features(tokens, index=0, stemmer=None, memory=2):
    """
    `tokens`  = a tokenized sentence [w1, w2, ...]
    `index`   = the index of the token we want to extract features for
    """
    if isinstance(tokens, str):
        tokens = [tokens]
    original_toks = list(tokens)
    # init the stemmer
    stemmer = stemmer or SnowballStemmer('english')

    # Pad the sequence with placeholders
    tokens = []
    for i in range(1, memory + 1):
        tokens.append(f'__START{i}__')
    tokens = list(reversed(tokens)) + original_toks
    for i in range(1, memory + 1):
        tokens.append(f'__END{i}__')

    # shift the index to accommodate the padding
    index += memory

    word = tokens[index]
    feat_dict = extract_single_word_features(word)
    feat_dict["word"] = word
    feat_dict["shape"] = get_word_shape(word)
    feat_dict["lemma"] = stemmer.stem(word)

    # look ahead N words
    for i in range(1, memory + 1):
        k = "next-" * i
        nextword = tokens[index + i]
        feat_dict[k + "word"] = nextword
        feat_dict[k + "lemma"] = stemmer.stem(nextword)
        feat_dict[k + "shape"] = get_word_shape(nextword)

    # look back N words
    for i in range(1, memory + 1):
        k = "prev-" * i
        prevword = tokens[index - i]
        feat_dict[k + "word"] = prevword
        feat_dict[k + "lemma"] = stemmer.stem(prevword)
        feat_dict[k + "shape"] = get_word_shape(prevword)

    return feat_dict


def extract_single_word_features(word):

    feat_dict = {
        'suffix1': word[-1:],
        'suffix2': word[-2:],
        'suffix3': word[-3:],
        'prefix1': word[:1],
        'prefix2': word[:2],
        'prefix3': word[:3]
    }
    return feat_dict


def extract_rte_features(rtepair):
    extractor = nltk.RTEFeatureExtractor(rtepair)
    features = {}
    features['word_overlap'] = len(extractor.overlap('word'))
    features['word_hyp_extra'] = len(extractor.hyp_extra('word'))
    features['ne_overlap'] = len(extractor.overlap('ne'))
    features['ne_hyp_extra'] = len(extractor.hyp_extra('ne'))
    return features


from nltk.corpus import wordnet, names
from nltk.probability import FreqDist
from nltk.tag import SequentialBackoffTagger


class WordNetTagger(SequentialBackoffTagger):
    '''
    >>> wt = WordNetTagger()
    >>> wt.tag(['food', 'is', 'great'])
    [('food', 'NN'), ('is', 'VB'), ('great', 'JJ')]
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.wordnet_tag_map = {
            'n': 'NN',
            's': 'JJ',
            'a': 'JJ',
            'r': 'RB',
            'v': 'VB'
        }

    def choose_tag(self, tokens, index, history):
        word = tokens[index]
        fd = FreqDist()

        for synset in wordnet.synsets(word):
            fd[synset.pos()] += 1

        if not fd: return None
        return self.wordnet_tag_map.get(fd.max())


class NamesTagger(SequentialBackoffTagger):
    '''
    >>> nt = NamesTagger()
    >>> nt.tag(['Jacob'])
    [('Jacob', 'NNP')]
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name_set = set([n.lower() for n in names.words()])

    def choose_tag(self, tokens, index, history):
        word = tokens[index]

        if word.lower() in self.name_set:
            return 'NNP'
        else:
            return None


class LocationsTagger(SequentialBackoffTagger):
    '''
    >>> nt = NamesTagger()
    >>> nt.tag(['Jacob'])
    [('Jacob', 'NNP')]
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name_set = set([n.lower() for n in names.words()])

    def choose_tag(self, tokens, index, history):
        word = tokens[index]

        if word.lower() in self.name_set:
            return 'NNP'
        else:
            return None



from setuptools import setup

setup(
    name='JarbasModelZoo',
    version='0.2.0a2post1',
    packages=['JarbasModelZoo',
              'JarbasModelZoo.features',
              "JarbasModelZoo.nltk_taggers",
              "JarbasModelZoo.nltk_chunkers"],
    url='https://github.com/OpenJarbas/ModelZoo',
    license='apache-2.0',
    author='jarbasai',
    include_package_data=True,
    author_email='jarbasai@mailfence.com',
    description='pretrained models (down)loader'
)
